from unittest import TestCase

from dependency_injection.data_structures.dependency_injection_exception import DependencyInjectionException
from dependency_injection.test_dependency_injection.injected_class1 import InjectedClass
from dependency_injection.test_dependency_injection.injected_class2 import InjectedClass2
from dependency_injection.test_dependency_injection.test_class_no_type_hint import TestClassNoTypeHint
from dependency_injection.test_dependency_injection.test_class_singlecall import TestClassSinglecall
from dependency_injection.test_dependency_injection.test_class_singleton import TestSingletonInjection


class TestInjection(TestCase):
    def test_injection_singleton(self) -> None:
        try:
            test_class_singleton: TestSingletonInjection = TestSingletonInjection('my_name', 'no_name', InjectedClass(), InjectedClass2())
        except Exception as ex:
            pass
        ret_val: str = test_class_singleton.boo()
        self.assertEqual('hello TestClassSingleton from InjectedClass1', ret_val)

        try:
            test_class_singleton: TestSingletonInjection = TestSingletonInjection('my_name', 'no_name')
        except Exception as ex:
            pass
        ret_val: str = test_class_singleton.boo()
        self.assertEqual('hello TestClassSingleton from InjectedClass1', ret_val)

    def test_injection_with_builtin_types_exception(self) -> None:
        try:
            # No injection all parameters are passed. all ok.
            test_class_no_hint_type: TestClassNoTypeHint = TestClassNoTypeHint('my_name', 'no_name', InjectedClass(), InjectedClass2())
        except DependencyInjectionException as ex:
            pass

        ret_val: str = test_class_no_hint_type.boo()
        self.assertEqual('hello TestClassNoTypeHint from InjectedClass1', ret_val)

        try:
            # No type hint for class InjectedClass2 in TestClassNoTypeHint __init__
            test_class_no_hint_type: TestClassNoTypeHint = TestClassNoTypeHint('my_name', 'no_name',)
        except DependencyInjectionException as ex:
            self.assertEqual(ex.exception_description, 'type hints must be used in order to inject a type. param_name=[injected_class2]')

    def test_injection_no_type_hint(self):
        try:
            test_class_no_type_hint: TestClassNoTypeHint = TestClassNoTypeHint('my_name')
        except DependencyInjectionException as ex:
            self.assertEqual(ex.exception_description, 'Cannot inject builtin types. param_type=[str] param_name=[no_name]')

    def test_injection_singlecall(self) -> None:
        try:
            test_class_singlecall: TestClassSinglecall = TestClassSinglecall('my_name', 'no_name', InjectedClass(), InjectedClass2())
        except Exception as ex:
            pass
        ret_val: str = test_class_singlecall.boo()
        self.assertEqual('hello TestClassSinglecall from InjectedClass1', ret_val)

        try:
            test_class_singlecall: TestClassSinglecall = TestClassSinglecall('my_name', 'no_name')
        except Exception as ex:
            pass
        ret_val: str = test_class_singlecall.boo()
        self.assertEqual('hello TestClassSinglecall from InjectedClass1', ret_val)





class DependencyInjectionException(Exception):

    def __init__(self, exception_id: int, exception_description: str):
        super().__init__(exception_description)
        self.exception_id: int = exception_id
        self.exception_description: str = exception_description





class InjectedClass:

    def __init__(self):
        self.message: str = ' from InjectedClass1'
        print(f'InjectedClass1 created - id=[{id(self)}]')

    def foo(self, str1: str) -> str:
        return str1 + self.message



class InjectedClass2:

    def __init__(self):
        self.boo: str = 'boo'
        print(f'InjectedClass2 created - id=[{id(self)}]')

    def foo(self, str1: str) -> str:
        return str1*2 + self.boo



from dependency_injection.decorators.autowired import autowired
from dependency_injection.decorators.autowired_enums import AutoWiredType
from dependency_injection.test_dependency_injection.injected_class1 import InjectedClass
from dependency_injection.test_dependency_injection.injected_class2 import InjectedClass2


class TestClassSinglecall:

    @autowired(AutoWiredType.SINGLECALL)
    def __init__(self, my_name: str, no_name: str, injected_class1: InjectedClass, injected_class2: InjectedClass2):
        self.injected_class1 = injected_class1
        self.my_name = my_name

    def boo(self) -> str:
        return self.injected_class1.foo('hello TestClassSinglecall')



import collections
import json
from typing import TypeVar


DictionaryAsObject = TypeVar('DictionaryAsObject', bound='DictionaryAsObject')
class DictionaryAsObject(collections.MutableMapping):

    # ToDo implement MutableMapping methods pop, popitem, clear, update, and setdefault they are inherited from parent

    def __init__(self, **kwargs):
        if kwargs:
            self.__dict__.update(**kwargs)

    def __getitem__(self, key):
        return getattr(self, str(key))

    def __setitem__(self, key, value):
        return setattr(self, str(key), value)

    def __contains__(self, key):
        return key in self.__dict__

    def __getattribute__(self, item):
        return object.__getattribute__(self, item)

    def __getattr__(self, attr):
        attr = str(attr)

        value = self.__dict__.get(str(attr), None)

        if value:
            return value

        return None

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, v) -> None:
        self.__dict__.__delitem__(v)

    def __iter__(self):
        return self.__dict__.__iter__()

    def __eq__(self, other):
        if isinstance(other, DictionaryAsObject):
            return self.__dict__ == other.__dict__
        elif isinstance(other, dict):
            return self.__dict__ == other

        return False

    def get(self, k, d=None):
        return self.__dict__.get(k, d)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __repr__(self):
        return str(self.__dict__)

    @classmethod
    def from_json(cls, json_data: str) -> DictionaryAsObject:
        return cls.from_dict(json.loads(json_data))

    @classmethod
    def from_dict(cls, dict_data: dict) -> DictionaryAsObject:
        obj_dict: DictionaryAsObject = DictionaryAsObject()

        for key, value in dict_data.items():
            if type(value) == dict:
                obj_dict[key] = cls.from_dict(value)
            else:
                obj_dict[key] = value

        return obj_dict



import inspect
from threading import Lock
from typing import Callable


class LockOnce:
    def __init__(self, func: Callable):

        eli = inspect.getmembers(func)

        self.func: Callable = func
        self.mutex: Lock = Lock()

    def __call__(self, *args, **kwargs):
        with self.mutex:
            self.func(self.func.__call__, *args, **kwargs)



from typing import Callable

from dependency_injection.design_patterns.singleton_handler import SingletonHandler


class Singleton(type):
    def __call__(cls, *args, **kwargs):
        cls_init_func: Callable = super()
        return SingletonHandler.add_singleton(cls, cls_init_func, *args, **kwargs)

    @classmethod
    def clear_instance(mcs):
        pass



from functools import wraps
from threading import Lock
mutex: Lock = Lock()


def synchronized():
    """
        Notice: currently does not support nested locking beware will result in a deadlock
    """
    def wrapper(f):
        @wraps(f)
        def inner_wrapper(*args, **kwargs):
            lock: Lock = mutex

            # Try to use the calling class lock if it exist
            try:
                if args[0].lock:
                    lock = args[0].lock
            except Exception as ex:
                pass

            with lock:
                return f(*args, **kwargs)
        return inner_wrapper
    return wrapper



import builtins
from functools import wraps

from dependency_injection.decorators.autowired_enums import AutoWiredType
from dependency_injection.decorators.autowired_handler import autowired_handler


def autowired(auto_wired_type=AutoWiredType.SINGLETON):
    def wrapper(func):
        @wraps(func)
        def inner_wrapper(*args, **kwargs):
            return autowired_handler.autowire(func, auto_wired_type, *args, **kwargs)
        return inner_wrapper
    return wrapper



import inspect
import builtins
from typing import Callable

from dependency_injection.decorators.autowired_enums import AutoWiredType, AutoWiredState
from dependency_injection.data_structures.dependency_injection_exception import DependencyInjectionException
from dependency_injection.design_patterns.singleton import Singleton
from dependency_injection.design_patterns.singleton_handler import SingletonHandler


class AutowiredHandler(metaclass=Singleton):
    def __init__(self):
        self.builtin_types = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]

    def autowire(self, func: Callable, auto_wired_type: AutoWiredType, *args, **kwargs) -> Callable:
        signature: inspect.Signature = inspect.signature(func)
        parameter: inspect.Parameter = None
        annotation: type = None
        instance: object = None
        args_index: int = -1
        auto_wired_state: AutoWiredState = AutoWiredState.IN_ARGS

        for key in signature.parameters:
            args_index += 1
            parameter = signature.parameters[key]
            annotation = parameter.annotation

            if len(args) == args_index:
                auto_wired_state = AutoWiredState.OUT_OF_ARGS

            if auto_wired_state == AutoWiredState.IN_ARGS:
                kwargs[key] = args[args_index]
                continue

            if auto_wired_state == AutoWiredState.OUT_OF_ARGS:
                if key in kwargs:
                    continue

                elif annotation == inspect._empty:
                    """
                        type hint must be used in order to inject a type using python-dependency-injection 
                    """
                    raise DependencyInjectionException(1, f'type hints must be used in order to inject a type. param_name=[{key}]')

                else:
                    if annotation in self.builtin_types:
                        raise DependencyInjectionException(1, f'Cannot inject builtin types. param_type=[{annotation.__name__}] param_name=[{key}]')

                    if auto_wired_type == AutoWiredType.SINGLECALL:
                        try:
                            kwargs[key] = annotation()
                        except Exception as ex:
                            raise DependencyInjectionException(1, str(ex) + f'Cannot create object - param_type=[{annotation.__name__}] param_name=[{key}] object should have empty constructor')

                    elif auto_wired_type == AutoWiredType.SINGLETON:
                        instance = SingletonHandler.get_singleton(annotation)
                        if instance is None:
                            raise DependencyInjectionException(1, f'Cannot create singleton - param_type=[{annotation.__name__}] param_name=[{key}] object should have empty constructor')
                        kwargs[key] = instance

        return func(**kwargs)


autowired_handler: AutowiredHandler = AutowiredHandler()



from threading import Lock
from typing import Callable


class SingletonHandler:
    _lock: Lock = Lock()
    _class_locks: dict = {}
    _singletons: dict = {}

    @classmethod
    def add_singleton(cls, required_class: type, init_func: Callable, *args, **kwargs) -> object:
        # get specific class lock to avoid dead lock when creating singleton inside singleton
        cls._lock.acquire()
        if required_class.__name__ not in cls._class_locks:
            cls._class_locks[required_class.__name__] = Lock()
        class_lock = cls._class_locks.get(required_class.__name__, None)
        cls._lock.release()

        # using the class lock access the in
        class_lock.acquire()
        if required_class not in cls._singletons:
            cls._singletons[required_class] = init_func.__call__(*args, **kwargs)
        instance = cls._singletons[required_class]
        class_lock.release()

        return instance

    @classmethod
    def get_singleton(cls, required_class: type) -> object:
        # get specific class lock to avoid dead lock when creating singleton inside singleton
        cls._lock.acquire()
        if required_class.__name__ not in cls._class_locks:
            cls._class_locks[required_class.__name__] = Lock()
        class_lock = cls._class_locks.get(required_class.__name__, None)
        cls._lock.release()

        # using the class lock access the in
        instance: object = None
        class_lock.acquire()
        if required_class in cls._singletons:
            instance = cls._singletons[required_class]
        else:
            try:
                instance = required_class()
                cls._singletons[required_class] = instance
            except Exception as ex:
                instance = None
        class_lock.release()

        return instance

    @classmethod
    def remove_singleton(cls, required_class: type) -> object:
        # get specific class lock to avoid dead lock when creating singleton inside singleton
        cls._lock.acquire()
        if required_class.__name__ not in cls._class_locks:
            cls._class_locks[required_class.__name__] = Lock()
        class_lock = cls._class_locks.get(required_class.__name__, None)
        cls._lock.release()

        # using the class lock access the in
        class_lock.acquire()
        if required_class not in cls._singletons:
            return None
        instance = cls._singletons.pop(required_class)
        class_lock.release()

        return instance






from dependency_injection.decorators.autowired import autowired
from dependency_injection.test_dependency_injection.injected_class1 import InjectedClass
from dependency_injection.test_dependency_injection.injected_class2 import InjectedClass2


class TestSingletonInjection:
    @autowired()
    def __init__(self, my_name: str, no_name: str, injected_class: InjectedClass, injected_class2: InjectedClass2):
        self.injected_class1 = injected_class
        self.my_name = my_name
        self.no_name = no_name
        self.injected_class2 = injected_class2

    def boo(self) -> str:
        return self.injected_class1.foo('hello TestClassSingleton')



from setuptools import setup, find_packages

with open('README.md','r') as readme_file:
    long_description: str = readme_file.read()

setup(
    name="python-dependency-injection",
    version="0.0.1",
    author="Eli Malka",
    author_email="eli.malka.mail@gmail.com",
    description="Python-dependency-injection is a simple yet powerful mini-framework for dependency injection in Python",
    license="MIT",
    keywords="python dependency injection",
    url="https://github.com/emalka/python-dependency-injection",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent"
    ],
    install_requiers=[]
)


from functools import wraps
from threading import Lock
mutex: Lock = Lock()


def component():
    """
        Notice: currently does not support nested locking beware will result in a deadlock
    """
    def wrapper(f):
        @wraps(f)
        def inner_wrapper(*args, **kwargs):
            lock: Lock = mutex

            # Try to use the calling class lock if it exist
            try:
                if args[0].lock:
                    lock = args[0].lock
            except Exception as ex:
                pass

            with lock:
                return f(*args, **kwargs)
        return inner_wrapper
    return wrapper

# class Component:
#     def __init__(self, my_class: type):
#
#         self.my_class = my_class
#
#
#     def __call__(self, *args, **kwargs):
#         with self.mutex:
#             self.func(self.func.__call__, *args, **kwargs)

